import graphene
from django.conf import settings
from graphene.relay import Node
from graphene_django.converter import convert_django_field, convert_field_to_string
from graphene_django.debug import DjangoDebug
from localized_fields.fields import LocalizedField

from .caluma_data_source import schema as data_source_schema
from .caluma_form import (
    historical_schema as form_historical_schema,
    schema as form_schema,
)
from .caluma_workflow import schema as workflow_schema

convert_django_field.register(LocalizedField, convert_field_to_string)

# The consolidated mutation is created procedurally here to avoid
# exporting it as `Mutation`, which can be confusing
_mutation = type(
    "Mutation",
    (form_schema.Mutation, workflow_schema.Mutation, graphene.ObjectType),
    {},
)

query_inherit_from = [
    form_schema.Query,
    workflow_schema.Query,
    data_source_schema.Query,
    graphene.ObjectType,
]

types = [
    form_schema.TextQuestion,
    form_schema.ChoiceQuestion,
    form_schema.MultipleChoiceQuestion,
    form_schema.DynamicChoiceQuestion,
    form_schema.DynamicMultipleChoiceQuestion,
    form_schema.TextareaQuestion,
    form_schema.FloatQuestion,
    form_schema.IntegerQuestion,
    form_schema.DateQuestion,
    form_schema.TableQuestion,
    form_schema.FormQuestion,
    form_schema.FileQuestion,
    form_schema.StaticQuestion,
    form_schema.StringAnswer,
    form_schema.ListAnswer,
    form_schema.IntegerAnswer,
    form_schema.FloatAnswer,
    form_schema.DateAnswer,
    form_schema.TableAnswer,
    form_schema.FileAnswer,
    form_schema.CalculatedFloatQuestion,
    form_schema.ActionButtonQuestion,
    workflow_schema.SimpleTask,
    workflow_schema.CompleteWorkflowFormTask,
    workflow_schema.CompleteTaskFormTask,
]

historical_types = [
    form_historical_schema.HistoricalStringAnswer,
    form_historical_schema.HistoricalListAnswer,
    form_historical_schema.HistoricalIntegerAnswer,
    form_historical_schema.HistoricalFloatAnswer,
    form_historical_schema.HistoricalDateAnswer,
    form_historical_schema.HistoricalTableAnswer,
    form_historical_schema.HistoricalFileAnswer,
]

if settings.ENABLE_HISTORICAL_API:
    types += historical_types
    query_inherit_from.append(form_historical_schema.Query)


class Query(*query_inherit_from):
    node = Node.Field()
    if settings.DEBUG:
        debug = graphene.Field(DjangoDebug, name="_debug")


schema = graphene.Schema(
    query=Query,
    mutation=_mutation,
    # TODO: define what app exposes what types
    types=types,
)



import datetime
import functools
import inspect

import pytest
import urllib3
from django.apps import apps
from django.core.cache import cache
from factory import Faker
from factory.base import FactoryMetaClass
from graphene import ResolveInfo
from minio import Minio
from minio.datatypes import Object as MinioStatObject
from pytest_factoryboy import register

from .caluma_core.faker import MultilangProvider
from .caluma_core.models import HistoricalRecords
from .caluma_form import factories as form_factories
from .caluma_logging import factories as logging_factories
from .caluma_user.models import AnonymousUser, OIDCUser
from .caluma_workflow import factories as workflow_factories
from .schema import schema

Faker.add_provider(MultilangProvider)


def register_module(module):
    for _, obj in inspect.getmembers(module):
        if isinstance(obj, FactoryMetaClass) and not obj._meta.abstract:
            register(obj)


register_module(form_factories)
register_module(logging_factories)
register_module(workflow_factories)


@pytest.fixture(scope="function", autouse=True)
def _autoclear_cache():
    cache.clear()


@pytest.fixture
def admin_groups():
    return ["admin"]


@pytest.fixture
def admin_user(settings, admin_groups):
    return OIDCUser(
        b"sometoken", {"sub": "admin", settings.OIDC_GROUPS_CLAIM: admin_groups}
    )


@pytest.fixture
def admin_request(rf, admin_user):
    request = rf.get("/graphql")
    request.user = admin_user
    return request


@pytest.fixture
def anonymous_request(rf):
    request = rf.get("/graphql")
    request.user = AnonymousUser()
    return request


@pytest.fixture
def info(anonymous_request):
    """Mock for GraphQL resolve info embedding django request as context."""
    return ResolveInfo(
        None,
        None,
        None,
        None,
        schema=None,
        fragments=None,
        root_value=None,
        operation=None,
        variable_values=None,
        context=anonymous_request,
    )


@pytest.fixture
def admin_info(admin_request):
    """Mock for GraphQL resolve info embedding authenticated django request as context."""
    return ResolveInfo(
        None,
        None,
        None,
        None,
        schema=None,
        fragments=None,
        root_value=None,
        operation=None,
        variable_values=None,
        context=admin_request,
    )


def simple_history_middleware(next, root, info, **args):
    """Mimick behaviour of simple_history.middleware.HistoryRequestMiddleware."""
    HistoricalRecords.thread.request = info.context
    return next(root, info, **args)


@pytest.fixture
def schema_executor(anonymous_request):
    return functools.partial(
        schema.execute,
        context_value=anonymous_request,
        middleware=[simple_history_middleware],
    )


@pytest.fixture
def admin_schema_executor(admin_request):
    return functools.partial(
        schema.execute,
        context_value=admin_request,
        middleware=[simple_history_middleware],
    )


@pytest.fixture
def minio_mock(mocker):
    def side_effect(bucket, object_name, expires):
        return f"http://minio/download-url/{object_name}"

    stat_response = MinioStatObject(
        # taken from a real-world minio stat() call
        bucket_name="caluma-media",
        object_name="a3d0429d-5400-47ac-9d02-124592302631_attack.wav",
        etag="5d41402abc4b2a76b9719d911017c592",
        size=8200,
        last_modified=datetime.datetime(
            2021, 3, 5, 15, 24, 33, tzinfo=datetime.timezone.utc
        ),
        content_type="application/pdf",
        metadata=urllib3._collections.HTTPHeaderDict(
            {
                "Accept-Ranges": "bytes",
                "Content-Length": "5",
                "Content-Security-Policy": "block-all-mixed-content",
                "Content-Type": "binary/octet-stream",
                "ETag": '"5d41402abc4b2a76b9719d911017c592"',
                "Last-Modified": "Fri, 05 Mar 2021 15:24:33 GMT",
                "Server": "MinIO",
                "Vary": "Origin",
                "X-Amz-Request-Id": "16697BAAD69D2214",
                "X-Xss-Protection": "1; mode=block",
                "Date": "Fri, 05 Mar 2021 15:25:15 GMT",
            }
        ),
        owner_id=None,
        owner_name=None,
        storage_class=None,
        version_id=None,
    )
    mocker.patch.object(Minio, "presigned_get_object")
    mocker.patch.object(Minio, "presigned_put_object")
    mocker.patch.object(Minio, "stat_object")
    mocker.patch.object(Minio, "bucket_exists")
    mocker.patch.object(Minio, "make_bucket")
    mocker.patch.object(Minio, "remove_object")
    mocker.patch.object(Minio, "copy_object")
    Minio.presigned_get_object.side_effect = side_effect
    Minio.presigned_put_object.return_value = "http://minio/upload-url"
    Minio.stat_object.return_value = stat_response
    Minio.bucket_exists.return_value = True
    return Minio


@pytest.fixture
def data_source_settings(settings):
    settings.DATA_SOURCE_CLASSES = [
        "caluma.caluma_data_source.tests.data_sources.MyDataSource"
    ]


@pytest.fixture
def history_mock(mocker):
    mocker.patch.object(HistoricalRecords, "create_historical_record")


@pytest.fixture
def simple_case(case_factory, document_factory, question_factory, answer_factory):
    question = question_factory()
    document = document_factory()
    answer_factory(document=document, question=question)
    case = case_factory(document=document)

    return case


@pytest.fixture
def permission_classes(settings):
    # TODO: replicate this for other globally-configured class lists
    # such as VISIBILITY_CLASSES and VALIDATION_CLASSES
    config = apps.get_app_config("caluma_core")

    old_permissions = settings.PERMISSION_CLASSES

    def set_permission_classes(class_names):
        settings.PERMISSION_CLASSES = class_names
        config.ready()

    yield set_permission_classes
    settings.PERMISSION_CLASSES = old_permissions
    config.ready()


@pytest.fixture
def form_and_document(
    db,
    form_factory,
    question_factory,
    form_question_factory,
    document_factory,
    answer_factory,
):
    """
    Return a factory for a form and document.

    The return value is a 4-tuple with the following values:
        (form, document, questions_dict, answers_dict)

    Optionally add a table question (and a row in the document),
    and a subform for added complexity.

    The slugs are named as follows:

    * form: top_form
       * question: top_question
       * question: table
           * row_form: row_form
               * question: column
       * question: form_question
           * sub_form: sub_form
               * question: sub_question
    """

    def factory(use_table=False, use_subform=False):
        form = form_factory(slug="top_form", meta={"is-top-form": True, "level": 0})
        document = document_factory(form=form)

        questions = {}
        answers = {}

        questions["top_question"] = question_factory(
            slug="top_question",
            type="text",
            is_required="true",
            is_hidden="false",
        )

        form_question_factory(form=form, question=questions["top_question"])
        answers["top_question"] = answer_factory(
            document=document, question=questions["top_question"]
        )

        if use_table:
            row_form = form_factory(slug="row_form")
            questions["table"] = question_factory(
                type="table",
                slug="table",
                row_form=row_form,
                is_required="true",
                is_hidden="false",
            )
            form_question_factory(form=form, question=questions["table"])
            questions["column"] = question_factory(
                type="float", slug="column", is_required="true", is_hidden="false"
            )
            form_question_factory(form=row_form, question=questions["column"])

            answers["table"] = answer_factory(
                document=document, question=questions["table"]
            )

            row_doc = document_factory(form=row_form, family=document)
            answers["column"] = answer_factory(
                document=row_doc, question=questions["column"]
            )
            answers["table"].documents.add(row_doc)

        if use_subform:
            sub_form = form_factory(
                slug="sub_form", meta={"is-top-form": False, "level": 1}
            )
            questions["form"] = question_factory(
                type="form",
                slug="form",
                sub_form=sub_form,
                is_required="true",
                is_hidden="false",
            )
            form_question_factory(form=form, question=questions["form"])
            questions["sub_question"] = question_factory(
                slug="sub_question",
                type="text",
                is_required="true",
                is_hidden="false",
            )
            form_question_factory(form=sub_form, question=questions["sub_question"])

            answers["sub_question"] = answer_factory(
                document=document, question=questions["sub_question"]
            )

        return (form, document, questions, answers)

    return factory


@pytest.fixture
def sorted_snapshot(snapshot):
    """Return a sortable snapshot that sorts a list in a nested data structure.

    The first arg (name) denotes the key/variable in the data structure which holds sortable list.
    The second arg (key) is passed to the builtin `sorted` function.
    """

    def _sorted(name, key=None):
        def _(data, path):
            if isinstance(data, list) and path[-1][0] == name:
                return sorted(data, key=key)
            return data

        return _

    def custom_snapshot(*args):
        return snapshot(matcher=_sorted(*args))

    return custom_snapshot



from uuid import UUID

import pytest
from django.utils import timezone

from .. import models


@pytest.mark.parametrize("question__type", [models.Question.TYPE_TEXT])
def test_history(db, question, document, schema_executor, admin_schema_executor):
    query = """
        mutation MyStringAnswer($input: SaveDocumentStringAnswerInput!) {
          saveDocumentStringAnswer (input: $input) {
            answer {
              id
            }
          }
        }
    """

    result = admin_schema_executor(
        query,
        variable_values={
            "input": {
                "question": str(question.pk),
                "value": "dolor",
                "document": str(document.pk),
            }
        },
    )
    assert not result.errors
    assert (
        models.Answer.history.get(value="dolor").history_user
        == admin_schema_executor.keywords["context_value"].user.username
    )

    query = """
            mutation MyStringAnswer($input: SaveDocumentStringAnswerInput!) {
              saveDocumentStringAnswer (input: $input) {
                answer {
                  id
                }
              }
            }
        """

    result = schema_executor(
        query,
        variable_values={
            "input": {
                "question": str(question.pk),
                "value": "sit",
                "document": str(document.pk),
            }
        },
    )
    assert not result.errors
    assert models.Answer.history.count() == 2
    history = models.Answer.history.all()
    assert (
        history[1].history_user
        == admin_schema_executor.keywords["context_value"].user.username
    )
    assert history[1].value == "dolor"

    assert history[0].history_user == "AnonymousUser"
    assert history[0].value == "sit"


def test_document_as_of(
    db,
    form_factory,
    form_question_factory,
    answer_factory,
    document_factory,
    snapshot,
    schema_executor,
    admin_schema_executor,
):
    f = form_factory(slug="root-form")
    document = document_factory(form=f, id=UUID("890ca108-d93d-4725-9066-7d0bddad8230"))

    q1 = form_question_factory(
        question__type=models.Question.TYPE_TEXT,
        question__slug="test_question1",
        form=f,
    )
    save_answer_query = """
        mutation saveAnswer($input: SaveDocumentStringAnswerInput!) {
          saveDocumentStringAnswer(input: $input) {
            clientMutationId
          }
        }
    """

    input = {
        "document": document.pk,
        "value": "first admin - revision 1",
        "question": q1.question.slug,
    }

    result = admin_schema_executor(save_answer_query, variable_values={"input": input})
    assert not result.errors
    timestamp1 = timezone.now()

    input["value"] = "second admin - revision 2 - not queried"
    result = admin_schema_executor(save_answer_query, variable_values={"input": input})
    assert not result.errors

    input["value"] = "first anon - revision 3"
    result = schema_executor(save_answer_query, variable_values={"input": input})
    assert not result.errors
    timestamp2 = timezone.now()

    input["value"] = "second anon - revision 4"
    result = schema_executor(save_answer_query, variable_values={"input": input})
    assert not result.errors
    timestamp3 = timezone.now()

    document.answers.get(question=q1.question).delete()

    historical_query = """
        query documentAsOf($id: ID!, $asOf: DateTime!) {
          documentAsOf (id: $id, asOf: $asOf) {
            meta
            documentId
            historicalAnswers (asOf: $asOf) {
              edges {
                node {
                  ...on HistoricalStringAnswer {
                    __typename
                    value
                    historyUserId
                  }
                }
              }
            }
          }
        }
    """

    variables = {"id": document.pk, "asOf": timestamp1}

    result = admin_schema_executor(historical_query, variable_values=variables)
    assert not result.errors
    snapshot.assert_match(result.data)

    variables["asOf"] = timestamp2
    result = admin_schema_executor(historical_query, variable_values=variables)
    assert not result.errors
    snapshot.assert_match(result.data)

    variables["asOf"] = timestamp3
    result = admin_schema_executor(historical_query, variable_values=variables)
    assert not result.errors
    snapshot.assert_match(result.data)

    variables["asOf"] = timezone.make_aware(timezone.datetime(1900, 9, 15))
    result = admin_schema_executor(historical_query, variable_values=variables)
    assert result.errors


def test_historical_file_answer(
    db,
    minio_mock,
    form_factory,
    document_factory,
    form_question_factory,
    schema_executor,
):
    f = form_factory(slug="root-form")
    document = document_factory(form=f)

    q1 = form_question_factory(
        question__type=models.Question.TYPE_FILE,
        question__slug="test_question1",
        form=f,
    )
    save_answer_query = """
            mutation saveAnswer($input: SaveDocumentFileAnswerInput!) {
              saveDocumentFileAnswer(input: $input) {
                clientMutationId
              }
            }
        """

    input = {
        "document": document.pk,
        "value": "my_file - rev 1",
        "question": q1.question.slug,
    }

    result = schema_executor(save_answer_query, variable_values={"input": input})
    assert not result.errors
    hist_file_1 = models.File.history.first()
    timestamp1 = timezone.now()

    input["value"] = "my_file - rev 2"
    result = schema_executor(save_answer_query, variable_values={"input": input})
    assert not result.errors
    hist_file_2 = models.File.history.first()
    timestamp2 = timezone.now()

    input["value"] = "my_file - rev 3"
    result = schema_executor(save_answer_query, variable_values={"input": input})
    assert not result.errors
    file3 = document.answers.get(question=q1.question).file

    minio_mock.copy_object.assert_called()
    minio_mock.remove_object.assert_called()

    historical_query = """
            query documentAsOf($id: ID!, $asOf: DateTime!) {
              documentAsOf (id: $id, asOf: $asOf) {
                historicalAnswers (asOf: $asOf) {
                  edges {
                    node {
                      ...on HistoricalFileAnswer {
                        __typename
                        historyUserId
                        value (asOf: $asOf) {
                          name
                          downloadUrl
                          historyUserId
                          historyType
                          historicalAnswer {
                            id
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
        """

    variables = {"id": document.pk, "asOf": timestamp1}
    result = schema_executor(historical_query, variable_values=variables)
    assert not result.errors
    assert (
        result.data["documentAsOf"]["historicalAnswers"]["edges"][0]["node"]["value"][
            "downloadUrl"
        ]
        == f"http://minio/download-url/{hist_file_1.pk}_{hist_file_1.name}"
    )

    variables["asOf"] = timestamp2
    result = schema_executor(historical_query, variable_values=variables)
    assert not result.errors
    assert (
        result.data["documentAsOf"]["historicalAnswers"]["edges"][0]["node"]["value"][
            "downloadUrl"
        ]
        == f"http://minio/download-url/{hist_file_2.pk}_{hist_file_2.name}"
    )

    # This is the newest revision, so the uuid in the downloadUri must be the one
    # from the actual file and not from the history record.
    variables["asOf"] = timezone.now()
    result = schema_executor(historical_query, variable_values=variables)
    assert not result.errors
    assert (
        result.data["documentAsOf"]["historicalAnswers"]["edges"][0]["node"]["value"][
            "downloadUrl"
        ]
        == f"http://minio/download-url/{file3.pk}_{file3.name}"
    )


def test_historical_table_answer(
    db,
    form_factory,
    document_factory,
    form_question_factory,
    answer_factory,
    answer_document_factory,
    schema_executor,
    snapshot,
):
    f = form_factory(slug="root-form")
    row_f = form_factory(slug="row-form")

    q_main = form_question_factory(
        question__type=models.Question.TYPE_TABLE,
        question__slug="test_table_question1",
        question__row_form=row_f,
        form=f,
    )
    main_document = document_factory(form=f)

    q_row = form_question_factory(
        question__type=models.Question.TYPE_TEXT,
        question__slug="test_row_question1",
        form=row_f,
    )
    row1_document = document_factory(form=row_f)
    answer_factory(
        question=q_row.question, document=row1_document, value="first row value"
    )

    row2_document = document_factory(form=row_f)
    answer = answer_factory(
        question=q_row.question, document=row2_document, value="second row value"
    )

    ad = answer_document_factory(
        answer__question=q_main.question,
        answer__document=main_document,
        document=row1_document,
        sort=0,
    )
    answer_document_factory(answer=ad.answer, document=row2_document, sort=1)

    timestamp_init = timezone.now()

    answer.delete()
    timestamp_2 = timezone.now()

    historical_query = """
        query documentAsOf($id: ID!, $asOf1: DateTime!, $asOf2: DateTime!) {
          d1: documentAsOf (id: $id, asOf: $asOf1) {
            historicalAnswers (asOf: $asOf1) {
              edges {
                node {
                  ...on HistoricalTableAnswer {
                    __typename
                    value (asOf: $asOf1) {
                      historicalAnswers (asOf: $asOf1) {
                        edges {
                          node {
                            ...on HistoricalStringAnswer {
                              value
                              historyType
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
          d2: documentAsOf (id: $id, asOf: $asOf2) {
            historicalAnswers (asOf: $asOf2) {
              edges {
                node {
                  ...on HistoricalTableAnswer {
                    __typename
                    value (asOf: $asOf2) {
                      historicalAnswers (asOf: $asOf2) {
                        edges {
                          node {
                            ...on HistoricalStringAnswer {
                              value
                              historyType
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
    """

    variables = {"id": main_document.pk, "asOf1": timestamp_init, "asOf2": timestamp_2}
    result = schema_executor(historical_query, variable_values=variables)
    assert not result.errors
    snapshot.assert_match(result.data)


def test_history_answer_type(
    db,
    answer_factory,
    form_question_factory,
    document_factory,
    form,
    schema_executor,
    admin_schema_executor,
):
    """Test that the answer type is resolved correctly through history.

    The answer type should resolve to the type of the question of the given time (as_of).
    Resolve type correctly even if the question type of a given question-slug changed.
    """
    document = document_factory(form=form)

    old_question = form_question_factory(
        question__type=models.Question.TYPE_TEXT, form=form
    ).question

    # create first answer type text
    query = """
        mutation saveAnswer($input: SaveDocumentStringAnswerInput!) {
          saveDocumentStringAnswer(input: $input) {
            clientMutationId
          }
        }
    """

    result = admin_schema_executor(
        query,
        variable_values={
            "input": {
                "question": str(old_question.pk),
                "value": "some text answer",
                "document": str(document.pk),
            }
        },
    )

    old_date = timezone.now()
    old_answer = models.Answer.objects.get()

    historical_answer = old_answer.history.first()
    historical_question = old_question.history.first()

    # delete old_question and create one with identical slug but different type
    old_question.delete()
    new_question = form_question_factory(
        form=form,
        question__slug=historical_question.slug,
        question__type=models.Question.TYPE_INTEGER,
    ).question

    query = """
        mutation MyIntegerAnswer($input: SaveDocumentIntegerAnswerInput!) {
          saveDocumentIntegerAnswer (input: $input) {
            answer {
              id
            }
          }
        }
    """

    # add new integer answer
    result = admin_schema_executor(
        query,
        variable_values={
            "input": {
                "question": str(new_question.pk),
                "value": 123,
                "document": str(document.pk),
            }
        },
    )
    assert not result.errors

    historical_query = """
        query documentAsOf($id: ID!, $asOf: DateTime!) {
          documentAsOf (id: $id, asOf: $asOf) {
            meta
            documentId
            historicalAnswers (asOf: $asOf) {
              edges {
                node {
                  __typename
                  ...on HistoricalStringAnswer {
                    stringValue: value
                  }
                  ...on HistoricalIntegerAnswer {
                    integerValue: value
                  }
                }
              }
            }
          }
        }
    """

    # old date resolves to the string value
    variables = {"id": document.pk, "asOf": old_date}
    result = admin_schema_executor(historical_query, variable_values=variables)

    assert not result.errors
    assert (
        result.data["documentAsOf"]["historicalAnswers"]["edges"][0]["node"][
            "stringValue"
        ]
        == historical_answer.value
    )
    assert (
        result.data["documentAsOf"]["historicalAnswers"]["edges"][0]["node"][
            "__typename"
        ]
        == "HistoricalStringAnswer"
    )

    # current date resolves to integer value
    variables = {"id": document.pk, "asOf": timezone.now()}
    result = admin_schema_executor(historical_query, variable_values=variables)

    assert not result.errors
    assert (
        result.data["documentAsOf"]["historicalAnswers"]["edges"][0]["node"][
            "integerValue"
        ]
        == 123
    )
    assert (
        result.data["documentAsOf"]["historicalAnswers"]["edges"][0]["node"][
            "__typename"
        ]
        == "HistoricalIntegerAnswer"
    )



# Generated by Django 2.2.10 on 2020-02-14 10:58

from django.db import migrations, models

fields = ("document", "question", "slug")


def remove_duplicates(apps, schema_editor):
    DynamicOption = apps.get_model("caluma_form", "DynamicOption")
    HistoricalDynamicOption = apps.get_model("caluma_form", "HistoricalDynamicOption")
    db_alias = schema_editor.connection.alias

    # get soon-to-be unique_together values from all records
    all_options = DynamicOption.objects.using(db_alias).values(*fields).order_by()

    # group by those values; annotate with max `modified_at`
    grouped = all_options.annotate(max_modified_at=models.Max("modified_at"))

    for group in grouped:
        to_delete = DynamicOption.objects.using(db_alias).filter(
            **{x: group[x] for x in fields}
        )
        to_delete = to_delete.exclude(modified_at=group["max_modified_at"])

        historical_to_delete = HistoricalDynamicOption.objects.using(db_alias).filter(
            id__in=to_delete.values_list("id", flat=True)
        )

        historical_to_delete.delete()
        to_delete.delete()


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0028_auto_20200210_0929")]

    operations = [
        migrations.RunPython(remove_duplicates, migrations.RunPython.noop),
        migrations.AlterUniqueTogether(
            name="dynamicoption", unique_together={("slug", "document", "question")}
        ),
        migrations.AlterUniqueTogether(
            name="historicaldynamicoption", unique_together=set()
        ),
    ]



from uuid import uuid4

from caluma.caluma_data_source.data_sources import BaseDataSource
from caluma.caluma_data_source.utils import data_source_cache


class MyDataSource(BaseDataSource):
    info = {"en": "Nice test data source", "de": "Schöne Datenquelle"}
    default = [1, 2, 3]

    @data_source_cache(timeout=3600)
    def get_data(self, user):
        return [
            1,
            (5.5,),
            "sdkj",
            ("value", "info"),
            ["something"],
            [
                "translated_value",
                {"en": "english description", "de": "deutsche Beschreibung"},
            ],
        ]

    @data_source_cache(timeout=60)
    def get_data_test_string(self, info):
        return "test string"

    @data_source_cache(timeout=60)
    def get_data_uuid(self, info):
        return str(uuid4())

    @data_source_cache(timeout=1)
    def get_data_expire(self, info):
        return str(uuid4())


class MyOtherDataSource(MyDataSource):
    def validate_answer_value(self, value, document, question, info):
        return "Test 123"


class MyFaultyDataSource(BaseDataSource):
    info = "Faulty test data source"
    default = None

    @data_source_cache(timeout=3600)
    def get_data(self, user):
        return "just a string"


class MyOtherFaultyDataSource(BaseDataSource):
    info = "Other faulty test data source"
    default = None

    @data_source_cache(timeout=3600)
    def get_data(self, user):
        return [["just", "some", "strings"]]


class MyBrokenDataSource(BaseDataSource):
    info = "Other faulty test data source"
    default = [1, 2, 3]

    @data_source_cache(timeout=3600)
    def get_data(self, user):
        raise Exception()


class MyOtherBrokenDataSource(BaseDataSource):
    info = "Other faulty test data source"
    default = None

    @data_source_cache(timeout=3600)
    def get_data(self, user):
        raise Exception()



import pytest


@pytest.mark.parametrize(
    "first,last,before,after,has_next,has_previous",
    [
        (1, None, None, None, True, False),
        (None, 1, None, None, False, True),
        (None, None, None, None, False, False),
        (None, None, None, "YXJyYXljb25uZWN0aW9uOjI=", False, True),
        (None, None, "YXJyYXljb25uZWN0aW9uOjI=", None, True, False),
    ],
)
def test_has_next_previous(
    db,
    first,
    last,
    before,
    after,
    has_next,
    has_previous,
    schema_executor,
    document_factory,
):
    document_factory.create_batch(5)

    query = """
        query AllDocumentsQuery ($first: Int, $last: Int, $before: String, $after: String) {
          allDocuments(first: $first, last: $last, before: $before, after: $after) {
            pageInfo {
              hasNextPage
              hasPreviousPage
            }
            edges {
              node {
                id
              }
            }
          }
        }
    """

    inp = {"first": first, "last": last, "before": before, "after": after}

    result = schema_executor(query, variable_values=inp)

    assert not result.errors
    assert result.data["allDocuments"]["pageInfo"]["hasNextPage"] == has_next
    assert result.data["allDocuments"]["pageInfo"]["hasPreviousPage"] == has_previous



from ..caluma_form.models import Document
from . import models
from .jexl import FlowJexl, GroupJexl


def get_group_jexl_structure(
    work_item_created_by_group, case, prev_work_item=None, context=None
):
    return {
        "case": {"created_by_group": case.created_by_group},
        "work_item": {"created_by_group": work_item_created_by_group},
        "prev_work_item": {
            "addressed_groups": prev_work_item.addressed_groups
            if prev_work_item
            else None,
            "controlling_groups": prev_work_item.controlling_groups
            if prev_work_item
            else None,
        },
        "context": {
            "addressed_groups": context.get("addressed_groups") if context else None,
            "controlling_groups": context.get("controlling_groups")
            if context
            else None,
        },
    }


def get_jexl_groups(
    jexl, task, case, work_item_created_by_user, prev_work_item=None, context=None
):
    if jexl:
        return GroupJexl(
            validation_context=get_group_jexl_structure(
                work_item_created_by_user.group, case, prev_work_item, context
            ),
            task=task,
            case=case,
            work_item_created_by_user=work_item_created_by_user,
            prev_work_item=prev_work_item,
            dynamic_context=context,
        ).evaluate(jexl)

    return []


def get_jexl_tasks(jexl, case, user, prev_work_item, context=None):
    if jexl:
        evaluated = FlowJexl(
            case=case, user=user, prev_work_item=prev_work_item, dynamic_context=context
        ).evaluate(jexl)

        if not isinstance(evaluated, list):
            evaluated = [evaluated]

        return evaluated

    return []


def create_work_items(tasks, case, user, prev_work_item=None, context: dict = None):
    work_items = []

    for task in tasks:
        controlling_groups = get_jexl_groups(
            task.control_groups,
            task,
            case,
            user,
            prev_work_item if prev_work_item else None,
            context,
        )
        addressed_groups = get_jexl_groups(
            task.address_groups,
            task,
            case,
            user,
            prev_work_item if prev_work_item else None,
            context,
        )
        if task.is_multiple_instance:
            # for multiple instance tasks, create one work item per adressed group
            work_item_groups = [[group] for group in addressed_groups]

            # make sure that at least one work item is always created, even if addressed_groups is empty
            if not work_item_groups:
                work_item_groups = [[]]
        else:
            # for regular tasks, multiple groups can be adressed to one work item
            work_item_groups = [addressed_groups]

        for groups in work_item_groups:
            work_items.append(
                models.WorkItem.objects.create(
                    addressed_groups=groups,
                    controlling_groups=controlling_groups,
                    task_id=task.pk,
                    deadline=task.calculate_deadline(),
                    document=Document.objects.create_document_for_task(task, user),
                    case=case,
                    status=models.WorkItem.STATUS_READY,
                    created_by_user=user.username,
                    created_by_group=user.group,
                    previous_work_item=prev_work_item if prev_work_item else None,
                )
            )

    return work_items
